from typing import Callable, Iterable
from time import time
from .utils import all_equal

def _trial(function: Callable, args: Iterable):
    start_time = time()
    result = function(*args)
    end_time = time()
    return {
        'result': result,
        'time': end_time - start_time
    }

def benchmark(
    functions: list[Callable],
    arguments: Iterable,
    validate: bool = False,
    max_iters: int = None,
    max_seconds: int = None,
):
    times = {function: list() for function in functions}
    start_time = time()
    for i, args in enumerate(arguments):
        results = dict()
        for function in functions:
            trial = _trial(function, args)
            results[function] = trial['result']
            times[function].append(trial['time'])
        if validate:
            if not all_equal(results.values()):
                raise AssertionError('Outputs not all equal')
        elapsed_time = time() - start_time
        if max_iters is not None and i+1 >= max_iters:
            break
        if max_seconds is not None and elapsed_time >= max_seconds:
            break
    return times


from typing import List
from decimal import Decimal
from math import ceil

def pow_lt(seq: List[int], k: Decimal) -> bool:
    """
    Given an arbitrarily long sequence of positive integers
    seq = [e₁, e₂, e₃, ..., eₙ] and a positive number k, returns True if and only if
    e₁^(e₂^(e₃^(···^eₙ))) < k.
    Inputs are not validated. That is, k is assumed to be positive and the
    elements of seq are assumed to be positive integers.
    Parameters
    ----------
    seq: list of positive integers.
    k: positive number.
    Returns
    -------
        True if and only if seq[0]^(seq[1]^(seq[2]^···)) < k.
    """
    if not len(seq): # if len(seq) == 0
        return 1 < k

    def _pow_lt(seq, k):
        if len(seq) == 1 or seq[0] == 1:
            return seq[0] < k
        if seq[1]*(seq[0].bit_length()-1) >= ceil(k).bit_length():
            return False
        l = Decimal(k).ln()/Decimal(seq[0]).ln() # high precision logarithm
        return _pow_lt(seq[1:], l) if l > 1 else False

    return _pow_lt(seq, k)

def pow_list(seq: List[Decimal]) -> Decimal:
    """
    Given an arbitrarily long sequence of numbers seq = [e₁, e₂, e₃, ..., eₙ],
    computes e₁^(e₂^(e₃^(···^eₙ))).
    The elements of seq are assumed to be numbers without validation.
    Parameters
    ----------
    seq: list of numbers.
    Returns
    -------
        seq[0]^(seq[1]^(seq[2]^···)).
    """
    l = len(seq)
    if not l: # if len(seq) == 0
        return 1
    elif l == 1: # if len(seq) == 1
        return seq[0]

    def _pow_list(seq):
        if seq[0] == 1:
            return 1
        if len(seq) == 2:
            return seq[0]**seq[1]
        return seq[0]**_pow_list(seq[1:])
    
    return _pow_list(seq)


from typing import Tuple

def ext_gcd(a: int, b: int) -> Tuple[int, int, int]:
    """
    Given nonnegative integers a and b, uses the extended Euclidean algorithm to
    compute gcd(a,b) and the coefficients of Bézout's identity -- integers x and
    y such that
        ax + by = gcd(a,b).
    
    Parameters
    ----------
    a, b: nonnegative integers.

    Returns
    -------
        gcd(a,b), x, y.
    """
    if a < 0 or b < 0:
        raise ValueError('a and b must be nonnegative integers')
    
    prev_r, r, prev_x, x = a, b, 1, 0
    while r: # while remainder > 0
        q = prev_r // r # quotient
        prev_r, prev_x, r, x = r, x, prev_r-q*r, prev_x-q*x
    return prev_r, prev_x, ((prev_r - prev_x*a)//b if b else 0)


from typing import List
from sympy.ntheory import totient
from .utils import pow_lt, pow_list

try:
    from gmpy2 import gcd, powmod, gcdext
except ImportError:
    from math import gcd
    powmod = pow
    from .gcd import ext_gcd as gcdext

def mod_nest_exp(seq: List[int], m: int) -> int:
    """
    Given an arbitrarily long sequence of positive integers
    seq = [a₁, a₂, a₃, ..., aₙ] and a positive integer m, computes
    a₁^(a₂^(a₃^(···^aₙ))) mod m.
    Parameters
    ----------
    seq: list of positive integers.
    m: positive integer
        The modulus of congruence.
    
    Returns
    -------
        seq[0]^(seq[1]^(seq[2]^···)) mod m.
    """
    if m < 1:
        raise ValueError('m must be a positive integer')
    for a in seq:
        if a < 1:
            raise ValueError('seq may only contain positive integers')

    if m == 1: # 1 divides every integer
        return 0
    l = len(seq)
    if not l: # if len(seq) == 0
        return 1%m
    elif l == 1: # if len(seq) == 1
        return seq[0]%m

    def _mod_nest_exp(seq, m):
        if m == 1: # 1 divides every integer
            return 0
        if len(seq) == 2: # recursive base case
            return powmod(seq[0], seq[1], m)
        
        b, e = seq[0], seq[1:] # base and exponent
        g = gcd(b, m)
        if g == 1:
            return powmod(b, _mod_nest_exp(e, totient(m)), m)
        
        n, k = m//g, 1
        g_ = gcd(g, n)
        while g_ > 1:
            n //= g_
            k += 1
            g_ = gcd(g, n)
        h = m//n
        _, x, y = gcdext(n, h)
        return (h*(y%n)*powmod(b, _mod_nest_exp(e, totient(n)), n)+
                n*(x%h)*(powmod(b, pow_list(e), h) if pow_lt(e, k) else 0))%m
    
    return _mod_nest_exp(seq, m)


import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='mod-nest-exp',
    version='1.0.3',
    author='Aviv Brook',
    author_email='avbrook@ucsc.edu',
    description='An algorithm that computes modular nested exponentiation efficiently',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/avivbrook/modular-nested-exponentiation',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Operating System :: OS Independent'
    ],
    package_dir={'': 'src'},
    install_requires=['gmpy2==2.0.8', 'sympy==1.7.1'],
    packages=setuptools.find_packages(where='src'),
    python_requires='>=3.6'
)